////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Project   :  -   Compound File
//  * URL       : http://alexander-bagel.blogspot.ru/2015/03/compound-file.html
//  * Unit Name : uRawReader
//  * Purpose   :     RAW 
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2015.
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.ru
//  ****************************************************************************
//

unit uRawReader;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, ExtCtrls, ImgList, FileCtrl;

type
  TForm2 = class(TForm)
    edSrc: TLabeledEdit;
    edDst: TLabeledEdit;
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    GroupBox1: TGroupBox;
    TreeView1: TTreeView;
    ImageList1: TImageList;
    OpenDialog1: TOpenDialog;
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form2: TForm2;

implementation

uses
  PoifsWithRepair,
  StorageTree;

{$R *.dfm}

// =============================================================================

procedure TForm2.Button2Click(Sender: TObject);
begin
  if OpenDialog1.Execute then
    edSrc.Text := OpenDialog1.FileName;
end;

procedure TForm2.Button3Click(Sender: TObject);
var
  S: string;
begin   {
  if SelectDirectory('Select directory to extract:', '', S,
    [sdNewFolder, sdNewUI, sdValidateDir]) then
    edDst.Text := S;}
end;

procedure TForm2.FormCreate(Sender: TObject);
begin
  edSrc.Text :=
    ExpandFileName(ExtractFilePath(ParamStr(0)) + '..\data\simple.bin');
  edDst.Text := ExpandFileName(ExtractFilePath(ParamStr(0)) + '..\data\simple\');
end;

procedure TForm2.Button1Click(Sender: TObject);
var
  FileStream: TFileStream;
  AFile: TPoifsFile;
  ATree: TStorageTree;
  I: Integer;
  DebugLog: TStringList;

  procedure FillAllChilds(RootIndex, CurrentIndex: Integer);
  var
    SubChildIndex: Integer;
    RootNode, CurrNode, ChildNode: TStorageElement;
  begin

    if CurrentIndex < 0 then Exit;

    //     
    RootNode := ATree.GetNode(RootIndex);

    //       
    CurrNode := ATree.GetNode(CurrentIndex);

    if CurrNode = nil then Exit;    

    //     - 
    if CurrNode.Added then Exit;
    CurrNode.Added := True;

    //      
    ATree.AddVector(RootNode, CurrNode);
    //        
    FillAllChilds(CurrNode.ID, CurrNode.Data.ChildProp);

    //         
    SubChildIndex := CurrNode.Data.PreviousProp;
    while SubChildIndex >= 0 do
    begin
      //  ,    
      FillAllChilds(RootIndex, SubChildIndex);
      ChildNode := ATree.GetNode(SubChildIndex);
      if ChildNode <> nil then
        SubChildIndex := ChildNode.Data.PreviousProp
      else
        SubChildIndex := -1;
    end;

    // -      ,     
    SubChildIndex := CurrNode.Data.NextProp;
    while SubChildIndex >= 0 do
    begin
      FillAllChilds(RootIndex, SubChildIndex);
      ChildNode := ATree.GetNode(SubChildIndex);
      if ChildNode <> nil then
        SubChildIndex := ChildNode.Data.NextProp
      else
        SubChildIndex := -1;
    end;
  end;

  procedure FillTree(Node: TTreeNode; RootNodeIndex: Integer);
  var
    W: WideString;
    TreeNode: TTreeNode;
    I: Integer;
    RootStorageNode, ChildStorageNode: TStorageElement;
  begin
    //   
    RootStorageNode := ATree.GetNode(RootNodeIndex);
    //     (   )
    W := RootStorageNode.Data.Caption;
    TreeNode := TreeView1.Items.AddChildFirst(Node, W);
    case RootStorageNode.Data.PropertyType of
      STGTY_STORAGE: TreeNode.ImageIndex := 0;
      STGTY_STREAM: TreeNode.ImageIndex := 1;
    end;
    //     
    for I := 0 to RootStorageNode.VectorCount - 1 do
    begin
      // ,       (   ?)
      ChildStorageNode := TStorageElement(RootStorageNode.GetVector(I).SlaveNode);
      if ChildStorageNode = nil then
        Continue;
      //   ,      ,   
      if ChildStorageNode.ID <> RootNodeIndex then
        FillTree(TreeNode, ChildStorageNode.ID);
    end;
  end;

  procedure GetStorageData(ANode: TStorageElement; const Stream: TStream);
  begin
    if ANode.Data.Size < Integer(AFile.Header._ulMiniSectorCutoff) then
      AFile.GetDataFromMiniStream(ANode.Data.StartBlock, ANode.Data.Size, Stream)
    else
      AFile.GetDataFromStream(ANode.Data.StartBlock, ANode.Data.Size, Stream);
  end;

  procedure Extract(Path: string; RootNodeIndex: Integer);
  var
    W: WideString;
    I: Integer;
    RootStorageNode, ChildStorageNode: TStorageElement;
    F: TFileStream;
  begin
    RootStorageNode := ATree.GetNode(RootNodeIndex);
    W := RootStorageNode.Data.Caption;
    case RootStorageNode.Data.PropertyType of
      STGTY_STORAGE:
        Path := Path + W + '\';
      STGTY_STREAM:
      begin
        try
          ForceDirectories(Path);
          F := TFileStream.Create(Path + W, fmCreate);
          try
            GetStorageData(RootStorageNode, F);
          finally
            F.Free;
          end;
        except
          DebugLog.Add(Path + W);
        end;
      end;
    end;
    for I := 0 to RootStorageNode.VectorCount - 1 do
    begin
      ChildStorageNode := TStorageElement(RootStorageNode.GetVector(I).SlaveNode);
      if ChildStorageNode = nil then
        Continue;
      if ChildStorageNode.ID <> RootNodeIndex then
        Extract(Path, ChildStorageNode.ID);
    end;
  end;

begin
  FileStream := TFileStream.Create(edSrc.Text, fmOpenReadWrite);
  try
    AFile := TPoifsFile.Create(FileStream);
    try
      //     
      AFile.LoadFromStream;

      ATree := TStorageTree.Create;
      try
        //  
        for I := 0 to AFile.PropertiesCount - 1 do
          ATree.AddNode(I).Data := AFile[I];

        //   
        FillAllChilds(0, ATree.GetNode(0).Data.ChildProp);

        //   
        TreeView1.Items.Clear;
        FillTree(nil, 0);

        //   
        DebugLog := TStringList.Create;
        try
          Extract(IncludeTrailingPathDelimiter(edDst.Text), 0);
          if DebugLog.Count > 0 then
            DebugLog.SaveToFile(IncludeTrailingPathDelimiter(edDst.Text) + 'cannotread.log');
        finally
          DebugLog.Free;
        end;
      finally
        ATree.Free;
      end;

    finally
      AFile.Free;
    end;
  finally
    FileStream.Free;
  end;
end;

end.
