////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Project   :  -   Compound File
//  * URL       : http://alexander-bagel.blogspot.ru/2015/03/compound-file.html
//  * Unit Name : uStgReader
//  * Purpose   :   
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2015.
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.ru
//  ****************************************************************************
//

unit uStgReader;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ExtCtrls, StdCtrls, FWStorage, ImgList;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    btnCreateDFase: TButton;
    btnAddFolder: TButton;
    btnAddFile: TButton;
    btnDelFolder: TButton;
    btnDelFile: TButton;
    btnOpenDBase: TButton;
    ListView1: TListView;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    ImageList1: TImageList;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnAddFolderClick(Sender: TObject);
    procedure ListView1DblClick(Sender: TObject);
    procedure ListView1SelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure btnDelFolderClick(Sender: TObject);
    procedure btnAddFileClick(Sender: TObject);
    procedure btnDelFileClick(Sender: TObject);
    procedure btnCreateDFaseClick(Sender: TObject);
    procedure btnOpenDBaseClick(Sender: TObject);
  private
    FCurrentFileName: string;
    FStorage: TFWStorage;
    FRoot: TFWStorageCursor;
    procedure OpenFile(CreateNew: Boolean);
    procedure ShowStorageData(AStorage: TFWStorageCursor);
    procedure EditFile;
  end;

var
  Form1: TForm1;

implementation

uses
  ActiveX,
  uEditFile;

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
begin
  //    
  FCurrentFileName :=
    ExpandFileName(ExtractFilePath(ParamStr(0)) + '..\data\simple.bin');
  //  
  FStorage := TFWStorage.Create;
  //  
  OpenFile(False);
end;

procedure TForm1.OpenFile(CreateNew: Boolean);
begin
  //  ,     
  FStorage.CloseFile;
  //   
  FStorage.OpenFile(FCurrentFileName,
    CreateNew or not FileExists(FCurrentFileName), FRoot);
  Caption := FCurrentFileName;
  //      
  ShowStorageData(FRoot);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FStorage.Free;
end;

procedure TForm1.ShowStorageData(AStorage: TFWStorageCursor);

  procedure AddItem(const ACaption: string; AIndex: Integer);
  begin
    with ListView1.Items.Add do
    begin
      Caption := ACaption;
      case AIndex of
        -1: ImageIndex := -1;
        1:
        begin
          ImageIndex := 0;
          SubItems.Add('Folder');
        end
      else
        ImageIndex := 1;
        SubItems.Add('File');
      end;
      //      Data, :
      // -1 -    
      //  0 - 
      //  1 - 
      //      
      Data := Pointer(AIndex);
    end;
  end;

var
  AData: TFWStorageEnum;
  I: Integer;
begin
  ListView1.Items.BeginUpdate;
  try
    ListView1.Items.Clear;
    //  ,        
    // (   - )
    if not AStorage.IsRoot then
      AddItem('..', -1);
    //    
    AStorage.Enumerate(AData);
    //     ListView
    for I := 0 to AData.Count - 1 do
      AddItem(
        string(AData.ElementEnum[I].pacsName),
        Byte(AData.ElementEnum[I].dwType = STGTY_STORAGE));
  finally
    ListView1.Items.EndUpdate;
  end;
end;

procedure TForm1.btnAddFolderClick(Sender: TObject);
var
  NewFolderName: string;
  Tmp: TFWStorageCursor;
begin
  if InputQuery('New folder', 'Enter folder name', NewFolderName) then
  begin
    FRoot.CreateStorage(AnsiString(NewFolderName), Tmp);
    FRoot.FlushBuffer;
  end;
  ShowStorageData(FRoot);
end;

procedure TForm1.ListView1DblClick(Sender: TObject);
begin
  //     - 
  if ListView1.Selected = nil then Exit;
  //    Data  
  case Integer(ListView1.Selected.Data) of
    -1: //    
    begin
      //      
      FRoot.Backward(FRoot);
      //   
      ShowStorageData(FRoot);
    end;
    0: //  
      EditFile;
    1: //  
    begin
      //     
      FRoot.OpenStorage(AnsiString(ListView1.Selected.Caption), FRoot);
      //   
      ShowStorageData(FRoot);
    end;
  end;
end;

procedure TForm1.ListView1SelectItem(Sender: TObject; Item: TListItem;
  Selected: Boolean);
begin
  if ListView1.Selected = nil then
  begin
    btnDelFolder.Enabled := False;
    btnDelFile.Enabled := False;
    Exit;
  end;
  btnDelFolder.Enabled := Integer(ListView1.Selected.Data) = 1;
  btnDelFile.Enabled := Integer(ListView1.Selected.Data) = 0;
end;

procedure TForm1.btnDelFolderClick(Sender: TObject);
begin
  if Application.MessageBox(
    PChar(Format('Delete folder: "%s"?', [ListView1.Selected.Caption])),
    'Delete folder', MB_ICONQUESTION or MB_YESNO) = ID_YES then
  begin
    FRoot.DeleteStorage(AnsiString(ListView1.Selected.Caption));
    FRoot.FlushBuffer;
    ShowStorageData(FRoot);
  end;
end;

procedure TForm1.btnAddFileClick(Sender: TObject);
var
  NewFileName: string;
begin
  if InputQuery('New file', 'Enter file name', NewFileName) then
  begin
    FRoot.CreateStream(AnsiString(NewFileName));
    FRoot.FlushBuffer;
  end;
  ShowStorageData(FRoot);
end;

procedure TForm1.btnDelFileClick(Sender: TObject);
begin
  if Application.MessageBox(
    PChar(Format('Delete file: "%s"?', [ListView1.Selected.Caption])),
    'Delete file', MB_ICONQUESTION or MB_YESNO) = ID_YES then
  begin
    FRoot.DeleteStream(AnsiString(ListView1.Selected.Caption));
    FRoot.FlushBuffer;
    ShowStorageData(FRoot);
  end;
end;

procedure TForm1.EditFile;
var
  Buff: TMemoryStream;
  Data: AnsiString;
begin
  Buff := TMemoryStream.Create;
  try
    //   
    FRoot.ReadStream(AnsiString(ListView1.Selected.Caption), Buff);
    //    
    if Buff.Size > 0 then
    begin
      SetLength(Data, Buff.Size);
      Buff.Read(Data[1], Buff.Size);
    end;
    //   
    frmEdit := TfrmEdit.Create(Self);
    try
      //   Memo  
      frmEdit.Memo1.Text := string(Data);
      //  
      if frmEdit.ShowModal <> mrOk then Exit;

      //    Memo
      Buff.Clear;
      Data := AnsiString(frmEdit.Memo1.Text);
      if Length(Data) > 0 then
        Buff.Write(Data[1], Length(Data));

      //   
      FRoot.WriteStream(AnsiString(ListView1.Selected.Caption), Buff);

      //   
      FRoot.FlushBuffer;
    finally
      frmEdit.Release;
    end;
  finally
    Buff.Free;
  end;
end;

procedure TForm1.btnCreateDFaseClick(Sender: TObject);
begin
  if SaveDialog1.Execute then
  begin
    FCurrentFileName := SaveDialog1.FileName;
    OpenFile(True);
  end;
end;

procedure TForm1.btnOpenDBaseClick(Sender: TObject);
begin
  if OpenDialog1.Execute then
  begin
    FCurrentFileName := OpenDialog1.FileName;
    OpenFile(False);
  end;
end;

end.
