////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Project   :  -   Compound File
//  * URL       : http://alexander-bagel.blogspot.ru/2015/03/compound-file.html
//  * Unit Name : StorageReadDemo
//  * Purpose   :   
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2015.
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.ru
//  ****************************************************************************
//

program StorageReadDemo;

{$APPTYPE CONSOLE}

uses
  Windows,
  SysUtils,
  ActiveX,
  AxCtrls;

procedure CheckHResult(Code: HRESULT);
begin
  if not Succeeded(Code) then
    RaiseLastOSError;
end;

procedure ShowFileData(AFile: IStream);
var
  OS: TOleStream;
  Buff: AnsiString;
begin
  Write('file data: ');
  OS := TOleStream.Create(AFile);
  try
    if OS.Size > 0 then
    begin
      SetLength(Buff, OS.Size);
      OS.ReadBuffer(Buff[1], OS.Size);
    end
    else
      Buff := 'file is empty';
    Writeln(Buff);
  finally
    OS.Free;
  end;
end;

procedure Enumerate(const Root: string; Storage: IStorage);
var
  Enum: IEnumStatStg;
  TmpElement: TStatStg;
  ShellMalloc: IMalloc;
  Fetched: Int64;
  Folder: IStorage;
  AFile: IStream;
begin
  // ..   OLE,     IMalloc
  if (CoGetMalloc(1, ShellMalloc) <> S_OK) or (ShellMalloc = nil) then
    raise Exception.Create('CoGetMalloc failed.');

  //       
  CheckHResult(Storage.EnumElements(0, nil, 0, Enum));

  //      
  Fetched := 1;
  while Fetched > 0 do
    if Enum.Next(1, TmpElement, @Fetched) = S_OK then
      //  ( )
      if ShellMalloc.DidAlloc(TmpElement.pwcsName) = 1 then
      begin
        //   
        Write('Found: ', Root, '\', AnsiString(TmpElement.pwcsName));
        //   
        case TmpElement.dwType of
          //   -      
          STGTY_STREAM:
          begin
            Writeln(' - file: ', sLineBreak);
            CheckHResult(Storage.OpenStream(TmpElement.pwcsName, nil,
              STGM_READ or STGM_SHARE_EXCLUSIVE, 0, AFile));
            ShowFileData(AFile);
            Writeln;
          end;
          //   -          
          STGTY_STORAGE:
          begin
            Writeln(' - folder');
            CheckHResult(Storage.OpenStorage(TmpElement.pwcsName, nil,
              STGM_READ or STGM_SHARE_EXCLUSIVE, nil, 0, Folder));
            Enumerate(Root + '\' + string(TmpElement.pwcsName), Folder);
          end;
        else
          Writeln('Unsupported type: ', TmpElement.dwType);
        end;
        // ,      -   
        ShellMalloc.Free(TmpElement.pwcsName);
      end;
end;

var
  TestFilePath: string;
  WideBuff: WideString;
  Root: IStorage;
begin
  TestFilePath := ExtractFilePath(ParamStr(0)) + '..\data\simple.bin';

  //    
  WideBuff := TestFilePath;
  CheckHResult(StgOpenStorage(@WideBuff[1], nil,
    STGM_READ or STGM_SHARE_DENY_WRITE, nil, 0, Root));

  //          
  Enumerate('root', Root);

  Writeln('DONE');
  Readln;
end.
